#include "main.h"

void 
recv_client_message(int index)
{
  int len, i, j;

  gs.client_recv_buffer[0] = '\0';
  if( (len = recv(gs.client[index].socket, gs.client_recv_buffer, 
                  sizeof(gs.client_recv_buffer) - 1, 0)) < 0 ) 
  {
      print_to_log("Recv error from client: %s\n", strerror(errno));
      disconnect_client(index);
      return;
  }

  if( strlen(gs.client_recv_buffer) == 0 ) {
    disconnect_client(index);
  } else {
    for(i=0; i<len; i++) {
      if( gs.client_recv_buffer[i] == '\n' || gs.client_recv_buffer[i] == '\r' ) {
        gs.client_recv_buffer[i] = '\0';
      }
    }
    i = 0;
    while( i<len ) {
      j = i; 
      i += strlen(&gs.client_recv_buffer[i]) + 1;
      process_client_command(&gs.client_recv_buffer[j], index);
    }
  }
}

void 
process_client_command(char *text, int index)
{
  char  cmd[MAX_STRING_LEN], str1[MAX_STRING_LEN];
  char  rest1[MAX_STRING_LEN], rest2[MAX_STRING_LEN], rest3[MAX_STRING_LEN];
  int   i, num1, num2, sentFlag = FALSE;

  if( debug_check_mode(DEBUG_INCOMING_TCP_DATA) ) {
    print_to_log("TCPIN %s: %s\n", gs.client[index].realname, text);
  }

  if( strlen(text) >= MAX_STRING_LEN ) {
    print_to_log("Command too long!: %s\n", text);
    return;
  }

  if( sscanf(text, "%s",                   cmd) != 1 ) return;
  if( sscanf(text, "%*s %s",              str1) != 1 ) str1[0]  = '\0';
  if( sscanf(text, "%*s %[^\n]",         rest1) != 1 ) rest1[0] = '\0';
  if( sscanf(text, "%*s %*s %[^\n]",     rest2) != 1 ) rest2[0] = '\0';
  if( sscanf(text, "%*s %*d %[^\n]",     rest3) != 1 ) rest3[0] = '\0';
  if( sscanf(text, "%*s %d",             &num1) != 1 ) num1     =  -1;
  if( sscanf(text, "%*s %*d %d",         &num2) != 1 ) num2     =  -1;

  if(        strcmp(cmd, "password") == 0 ) {
      check_client_password(index, rest1);
  } else if( strcmp(cmd, "disconnect") == 0 ) {
      disconnect_client(index);
  } else if( gs.client[index].password == 0 ) {
      send_to_client(index, "commandprint No password, command ignored\n");
      return;
  } else if( strcmp(cmd, "masquerade") == 0 ) {
      sentFlag = TRUE;
      num1--;
      if( num1 >=0 && num1 < MAX_PLAYERS && gs.players[num1].valid ) {
        send_name_command(gs.players[num1].real_name);
        map_control_chars(rest2, rest3, MAX_STRING_LEN);
        send_console_command("say \"%s\"", rest2);
        send_name_command(gs.player_name);
      }
  } else if( strcmp(cmd, "say") == 0 ) {
      sentFlag = TRUE;
      send_name_command(gs.client[index].stagename);
      map_control_chars(rest2, rest1, MAX_STRING_LEN);
      send_console_command("say \"%s\"", rest2);
      send_name_command(gs.player_name);
  } else if( strcmp(cmd, "say_team") == 0 ) {
      sentFlag = TRUE;
      send_name_command(gs.client[index].stagename);
      map_control_chars(rest2, rest1, MAX_STRING_LEN);
      send_console_command("say_team \"%s\"", rest2);
      send_name_command(gs.player_name);
  } else if( strcmp(cmd, "say_qsmack") == 0 ) {
      map_control_chars(rest2, rest1, MAX_STRING_LEN);
      send_to_clients("commandprint #%-2d %-15s: %s\n", gs.player_number,
                   gs.client[index].stagename, rest2);
  } else if( strcmp(cmd, "servername") == 0 ) {
      map_control_chars(gs.player_name, rest1, MAX_NAMELEN);
      send_name_command(gs.player_name);
      send_to_clients("myname %s\n", map_weird_chars(gs.player_name));
  } else if( strcmp(cmd, "stagename") == 0 ) {
      map_control_chars(gs.client[index].stagename, rest1, MAX_NAMELEN);
      send_to_client(index, "mystagename %s\n", map_weird_chars(gs.client[index].stagename));
  } else if( strcmp(cmd, "realname") == 0 ) {
      strcpy(gs.client[index].realname, rest1);
      gs.client[index].valid = 0;
      send_to_clients("commandprint New QSmack Client '%s' has connected\n", 
                                 gs.client[index].realname);
      gs.client[index].valid = 1;
  } else if( strcmp(cmd, "status") == 0 ) {
      send_to_client(index, "commandprint Current player's status:\n");
      for(i=0; i<MAX_PLAYERS; i++) {
        if( gs.players[i].valid ) {
          send_to_client(index, "commandprint #%-2d %-15s %3d %2d:%-2d %-15s %9s\n", 
                         i+1,
                         gs.players[i].name,
                         gs.players[i].frags,
                         gs.players[i].shirt,
                         gs.players[i].pants,
                         gs.players[i].ipaddr,
                         gs.players[i].time);
        }
      }
  } else if( strcmp(cmd, "who") == 0 ) {
      send_to_client(index, "commandprint All people currently logged in:\n");
      for(i=0; i<MAX_CLIENTS; i++) {
        if( gs.client[i].valid ) {
          send_to_client(index, "commandprint %-40s stage name: %s, connected from: %s\n", 
               gs.client[i].realname, map_weird_chars(gs.client[i].stagename),
               net_address_to_name(&gs.client[i].addr));
        }
      }
  } else if( strcmp(cmd, "tell") == 0 ) {
      if( num1 != -1 ) {
        sentFlag = TRUE;
        map_control_chars(rest2, rest3, MAX_STRING_LEN);
        send_console_command("tell \"%s\" %s", gs.players[num1-1].real_name, rest2);
      }
  } else if( strcmp(cmd, "color") == 0 ) {
      if( num1 != -1 ) {
        gs.player_shirt = num1;
        gs.player_pants = (num2 != -1) ? num2 : gs.player_shirt;
        send_console_command("color %d %d", gs.player_shirt, gs.player_pants);
        send_to_clients("myshirt %d\n", gs.player_shirt);
        send_to_clients("mypants %d\n", gs.player_pants);
      }
  } else if( strcmp(cmd, "kick") == 0 ) {
      if( num1 != -1 && num1 != gs.player_number ) {
        sentFlag = TRUE;
        if( gs.kick_impulse >= 0 ) {
          send_impulse(num1 + gs.kick_impulse);
        } else if( gs.ban_impulse >= 0 ) {
          send_impulse(num1 + gs.ban_impulse);
        }
      }
  } else if( strcmp(cmd, "impulse") == 0 ) {
      if( num1 != -1 ) {
        send_impulse(num1);
      }
  } else if( strcmp(cmd, "fire") == 0 ) {
      sentFlag = TRUE;
      send_fire();
  } else if( strcmp(cmd, "jump") == 0 ) {
      sentFlag = TRUE;
      send_jump();
  } else if( strcmp(cmd, "quake") == 0 ) {
      sentFlag = TRUE;
      send_console_command("%s", rest1);
  } else if( strcmp(cmd, "banlist") == 0 ) {
      send_ban_list(index);
  } else if( strcmp(cmd, "unban") == 0 ) {
      if( num1 != -1 ) {
        remove_ban_item_by_num(num1);
        write_ban_file();
      }
  } else if( strcmp(cmd, "ban_ip") == 0 ) {
      if( strlen(str1) ) {
        add_ban_ip_item(str1, gs.client[index].realname, rest2); 
        write_ban_file();
      }
  } else if( strcmp(cmd, "ban_exclude") == 0 ) {
      if( strlen(str1) ) {
        add_ban_exclude_item(str1, gs.client[index].realname, rest2); 
        write_ban_file();
      }
  } else if( strcmp(cmd, "ban_name") == 0 ) {
      if( strlen(str1) ) {
        add_ban_name_item(str1, gs.client[index].realname, rest2);
        write_ban_file();
      }
  } else if( strcmp(cmd, "ban_color") == 0 ) {
      if( num1 != -1 ) {
        if( num2 == -1 ) num2 = num1;
        add_ban_color_item(num1, num2);
        write_ban_file();
      }
  } else if( strcmp(cmd, "reread_banfile") == 0 ) {
      destroy_ban_lists();
      read_ban_file(); 
  } else if( strcmp(cmd, "say_banlist") == 0 ) {
      say_ban_list();
  } else if( strcmp(cmd, "lockban") == 0 ) {
      gs.ban_lock = TRUE;
  } else if( strcmp(cmd, "unlockban") == 0 ) {
      gs.ban_lock = FALSE;
  } else if( strcmp(cmd, "qsmack_quit") == 0 ) {
      shutdown_all();
  } else if( strcmp(cmd, "qsmack_connect") == 0 ) {
      if( strlen(str1) ) {
        strcpy(gs.server_name, str1);
        connect_quake_server();
      }
  } else if( strcmp(cmd, "qsmack_disconnect") == 0 ) {
      sentFlag = TRUE;
      send_disconnect();
  } else {
      send_to_client(index, "exception Unknown command '%s' sent to qsmack server\n", cmd);
      print_to_log("Unknown command '%s' sent to qsmack server by %s\n", 
            cmd, gs.client[index].realname);
  }

  if( sentFlag == TRUE && !is_connected() ) {
    send_to_client(index, "commandprint Command: '%s' ignored, "
                          "QSmack not connected to quake server\n", cmd);
  }
}

void 
init_client_connections(void)
{
  int i;

  for(i=0; i<MAX_CLIENTS; i++) {
    gs.client[i].valid = 0;
    gs.client[i].password = 0;
    gs.client[i].realname[0] = '\0';
    gs.client[i].stagename[0] = '\0';
  }
}

void 
open_listen_connection(void)
{
  int sock_flag = 1;

  gs.listen_socket = net_open_socket(SOCK_STREAM);
  setsockopt(gs.listen_socket, SOL_SOCKET, SO_REUSEADDR, (char *)&sock_flag, 
             sizeof(int));
           
  memset((unsigned char *)&gs.listen_addr, 0, sizeof(gs.listen_addr));
  gs.listen_addr.sin_family       = AF_INET;
  gs.listen_addr.sin_addr.s_addr  = htonl(INADDR_ANY);
  net_bind_socket(gs.listen_socket, &gs.listen_addr, (short)gs.listen_port);
  net_listen(gs.listen_socket);

  print_to_log("Listening for connections on port %d\n", gs.listen_port);
}

void 
accept_new_client(void)
{
  struct sockaddr_in client_addr;
  int                client_addrlen = sizeof(client_addr);
  int                client_socket, i;

  client_socket = accept(gs.listen_socket, (struct sockaddr *)&client_addr, &client_addrlen);

  if( client_socket == -1 ) {
    print_to_log("Accept error from %s: %s\n", 
             net_address_to_name(&client_addr), strerror(errno));
    return;
  }

  for(i=0; i<MAX_CLIENTS; i++) {
    if( gs.client[i].valid == 0 ) {
      break;
    }
  }
 
  if( i == MAX_CLIENTS ) {
    send(client_socket, "rejected Server Full\n", 22, 0);
    net_close_socket(client_socket);
    print_to_log("Server Full: Refused connection to %s\n", 
                                       net_address_to_name(&client_addr));
  } else {
    gs.client[i].valid      = 1;
    gs.client[i].password   = 0;
    gs.client[i].socket     = client_socket;
    restart_timer(&gs.client[i].password_timer);
    strcpy(gs.client[i].realname,  net_address_to_name(&client_addr));
    memcpy(&gs.client[i].addr, &client_addr, client_addrlen);
    net_open_fd(gs.client[i].socket);
    print_to_log("Accepted connection from %s\n", 
                                 net_address_to_name(&gs.client[i].addr));
    send_to_client(i, "password\n");
  }
}

void 
check_client_password(int index, char *password)
{
  if( gs.password && strcmp(gs.password, password) == 0 ) {
    gs.client[index].password = 1;
    reset_timer(&gs.client[index].password_timer);
    send_to_client(index, "accepted\n");
    send_client_init_game_info(index);
  } else {
    send_to_client(index, "commandprint Invalid password\n");
    print_to_log("Invalid passoword, disconnecting %s\n", 
                     net_address_to_name(&gs.client[index].addr));
    disconnect_client(index);
  }
}

void 
send_client_init_game_info(int index)
{
  int                   i;
  struct stuff_command *sc = gs.stuff_commands;

  if( !is_in_the_game() ) {
    send_to_client(index, 
      "commandprint QSmack server is not connected to quake server %s\n",
      gs.server_name);
    send_to_client(index,
      "commandprint See server admin for further instructions\n");
  } 

  send_to_client(index, "myname %s\n",     map_weird_chars(gs.player_name));
  send_to_client(index, "myshirt %d\n",    gs.player_shirt);
  send_to_client(index, "mypants %d\n",    gs.player_pants);
  send_to_client(index, "mynumber %d\n",   gs.player_number);

  send_to_client(index, "server %s\n",     gs.server_name);
  send_to_client(index, "port %d\n",       gs.server_port);
  send_to_client(index, "mapname %s\n",    gs.mapname);
  send_to_client(index, "maxplayers %d\n", gs.max_players);

  for(i=0; i<MAX_PLAYERS; i++) {
    if( gs.players[i].valid ) {
      send_to_client(index, "playerip %d %s\n",    i, gs.players[i].ipaddr);
      send_to_client(index, "playername %d %s\n",  i, gs.players[i].name);
      send_to_client(index, "playername_raw %d %s\n", i, gs.players[i].raw_mapped_name);
      send_to_client(index, "playerfrags %d %d\n", i, gs.players[i].frags);
      send_to_client(index, "playershirt %d %d\n", i, gs.players[i].shirt);
      send_to_client(index, "playerpants %d %d\n", i, gs.players[i].pants);
    }
  }

  while( sc != NULL ) {
    send_to_client(index, "stuff %s", sc->command);
    sc = sc->next;
  }
}

void 
disconnect_all_clients(void)
{
  int i;

  for(i=0; i<MAX_CLIENTS; i++) {
    if( gs.client[i].valid ) {
      disconnect_client(i);
    }
  }
}

void 
disconnect_client(int index)
{
  gs.client[index].valid = 0;
  gs.client[index].password = 0;
  net_close_socket(gs.client[index].socket);
  print_to_log("Disconnecting client: %s\n", gs.client[index].realname);
  send_to_clients("commandprint QSmack Client \'%s\' has disconnected\n", 
               gs.client[index].realname);
  gs.client[index].realname[0] = '\0';
  gs.client[index].stagename[0] = '\0';
}

void 
send_to_clients(char *fmt, ...)
{
  va_list arglist;
  int     i;
  char    tmp[512];

  va_start(arglist, fmt);

  if( debug_check_mode(DEBUG_OUTGOING_TCP_DATA) ) {
    vsprintf(tmp, fmt, arglist);
    print_to_log("TCPOUT all: %s", tmp);
  }

  for(i=0; i<MAX_CLIENTS; i++) {
    if( gs.client[i].valid && gs.client[i].password ) {
      _send_to_client(i, fmt, arglist);
    }
  }

  va_end(arglist);
}

void 
send_to_client(int index, char *fmt, ...) 
{
  va_list arglist;
  char    tmp[512];
  
  va_start(arglist, fmt);
  if( debug_check_mode(DEBUG_OUTGOING_TCP_DATA) ) {
    vsprintf(tmp, fmt, arglist);
    print_to_log("TCPOUT %s: %s", gs.client[index].realname, tmp);
  }
  _send_to_client(index, fmt, arglist);
  va_end(arglist);
}

void 
_send_to_client(int index, char *fmt, va_list arglist) 
{
  int len = vsprintf(gs.client_send_buffer, fmt, arglist);

  if( send(gs.client[index].socket, gs.client_send_buffer, len, 0) != len ) {
    if( errno == EPIPE ) {
      disconnect_client(index);
    } else {
      print_to_log("Send error to client: %s\n", strerror(errno));
      disconnect_client(index);
    }
  }
}
