/*
 *  gstate.h
 *
 *  Header file for game state routines
 *
 *  Kekoa Proudfoot 1/4/97
 *
 */

#ifndef GSTATE_H_DEFINED
#define GSTATE_H_DEFINED

/* These should be defined in some qtypes.h file */

#define GSTATE_MAX_NAME_LENGTH       16
#define GSTATE_NUM_MODEL_NAMES       256
#define GSTATE_NUM_SOUND_NAMES       256
#define GSTATE_NUM_STATIC_ENTITIES   128
#define GSTATE_NUM_DYNAMIC_ENTITIES  450
#define GSTATE_NUM_LIGHT_STYLES      64
#define GSTATE_NUM_PLAYERS           16

typedef struct vec_t {
    float       x;
    float       y;
    float       z;
} vec_t;

typedef struct rot_t {
    float       pitch;
    float       yaw;
    float       roll;
} rot_t;

/*****************************************************************************
 * Type definitions
 *****************************************************************************/

typedef struct entity_t {
    char        model;
    char        frame;
    char        colormap;
    char        skin;
    char        attack_state;  /* dynamic only */
    vec_t       position;      /* dynamic only */ /* purpose unknown */
    rot_t       orientation;
} entity_t;

typedef struct player_t {
    short       entity;

    char       *name;
    short       frags;
    char        shirt_color;
    char        pants_color;
} player_t;

typedef struct client_t {
    char        player;
    short       entity;

    float       view_z_offset;
    float       view_pitch_offset;
    vec_t       velocity;
    rot_t       orientation;
    long        items;
    char        bits;
    char        weapon_frame;
    char        weapon_model;
    char        armor;
    short       health;
    char        current_ammo;
    char        ammo_shells;
    char        ammo_nails;
    char        ammo_rockets;
    char        ammo_cells;
    char        current_weapon;
} client_t;

typedef struct sound_t {
    char        sound;
    vec_t       position;
    float       volume;
    char        attenuation;
    char        channel;      /* dynamic only */
    short       entity;       /* dynamic only */
} sound_t;

typedef struct particle_t {
    vec_t       position;
    vec_t       velocity;
    char        color;
    char        count;
} particle_t;

typedef struct damage_t {
    char        armor;
    char        health;
    vec_t       position;
} damage_t;

typedef struct tempent_t {
    char        type;
    short       entity;        /* which created this tempent */
    vec_t       position;
    vec_t       destination;
} tempent_t;

typedef struct gstate_t {

    /* Stuff which is mostly constant */

    char        signon_num;
    char       *map_name;
    char        multiplayer;

    char        paused;
    char        intermission;
    char        finale;
    char        cd_from;
    char        cd_to;
    char        sell_screen;

    short       monsters;
    short       secrets;

    short       num_model_names;
    char       *model_names[GSTATE_NUM_MODEL_NAMES];

    short       num_sound_names;
    char       *sound_names[GSTATE_NUM_SOUND_NAMES];

    short       num_light_styles;
    char       *light_styles[GSTATE_NUM_LIGHT_STYLES];

    short       num_static_entities;
    entity_t    static_entities[GSTATE_NUM_STATIC_ENTITIES];

    short       num_dynamic_defaults;
    entity_t    dynamic_defaults[GSTATE_NUM_DYNAMIC_ENTITIES];

    short       num_static_sounds;
    sound_t     static_sounds[GSTATE_NUM_STATIC_SOUNDS];

    char        num_players;
    player_t    players[GSTATE_NUM_PLAYERS];

    /* Stuff which mostly varies with every update */

    char        last_server_time;
    char        last_client_time;

    client_t    client_data;

    short       num_sounds;
    sound_t     sounds[GSTATE_NUM_SOUNDS];

    short       num_dynamic_entities;
    entity_t    dynamic_entities[GSTATE_NUM_DYNAMIC_ENTITIES];

    char        console_text[GSTATE_NUM_CONSOLE_CHARS];

    char        stuffed_text[GSTATE_NUM_STUFFED_CHARS];

    char        centered_text[GSTATE_NUM_CENTERED_CHARS];

    short       num_particles;
    particle_t  particles[GSTATE_NUM_PARTICLES];

    short       num_damages;
    damage_t    damages[GSTATE_NUM_DAMAGES];

    short       num_temp_entities;
    tempent_t   temp_entities[GSTATE_NUM_TEMP_ENTITIES];
} gstate_t;

extern gstate_t *gstate_create         (void);
extern void      gstate_destroy        (gstate_t *gs);
extern void      gstate_process_packet (gstate_t *gs, packet_t *p);

#endif
